# _*_ coding: iso-8859-1 _*_
###########################################
## Gnration du service client SCRIBE   ##
## pour la gestion du poste              ##
## licence GPL Eole                      ##
###########################################

import os
import traceback
# pour utilisation en non-compil
os.environ['path']='%s;%s'%(os.environ['path'], os.path.join(os.path.abspath('.'), 'dlls'))
import shutil
from twisted.spread import pb
from twisted.internet import reactor

import win32api as wa
import win32con as wc
import ntsecuritycon as ntsec

from logging.handlers import RotatingFileHandler
import logging

from eole import reg, option
from eole.vnc import Vnc
from eole.os_type import type_os, sp
from eole.perms import AdjustPrivilege
from eole.process import killProcName
from eole.lance_cmd import lancecmd
from eole.replace_win_vars import convert
from eole import lecteur
from eole import logon_service
from eole.pbgest import PBgest
from eole.log import SERVICE_LOG, FORMATTER

# utilisation du reacteur Windoz
reactortype = "win32"

# dfinition du service
nom = "servscribe"
libelle = "Service Scribe"
version = "3.2"


class servFactory(pb.PBServerFactory):
    def buildProtocol(self, addr):
        #il s'agit d'une liste, il faut faire un "eval"
        if str(addr.host) not in option.get_ip_access_list():
            logging.error('Connexion non autorisee de %s'%addr.host)
            return None
        return pb.PBServerFactory.buildProtocol(self, addr)

class Echoer(pb.Root):
    """Classe contenant toutes les fonctions distantes
       que le serveur Scribe peut appeler
    """

    def __init__(self, parent):
        """initialisation du fichier de log
        """
        self.log = parent.log
        self.serveur = parent.serveur
        self.nom_scribe = option.get_nom_scribe()
        self.user_port = int(option.get_port_utilisateur())
        # repertoire d'installation
        self.inst_path = option.get_inst_path()
        # repertoire d'installation vnc
        self.vnc_path = os.path.join(self.inst_path, 'ultravnc')
        # Variables d'environnement
        self.tmpdir = parent.tmpdir
        self.windir = parent.windir
        self.vnc = Vnc(self.vnc_path, self.tmpdir)
        self.nodrives, self.noview = None, None
        AdjustPrivilege([ntsec.SE_SHUTDOWN_NAME, ntsec.SE_BACKUP_NAME])
        
    #######################
    # Fonctions distantes #
    #######################
    def remote_bonjour(self, old=""):
        """fonction de test, old sert  pouvoir passer un argument supplmentaire (utilisation dans un callback)
        """
        self.log.info("Appel de la fonction remote_bonjour, old=%s"%old)
        return "** %s %s **"%(libelle, version)


    def remote_logon(self, logon_dict):
        """fonction appele lors de l'ouverture de session
        """
        e = logon_dict['esu_dict']
        msg = 'User=%s, ESU_GU=%s, Machine=%s, ESU_GM=%s'%(e['USERNAME'], e['ESU_GU'], wa.GetComputerName(), e['ESU_GM'])
        self.log.info('Ouverture de session %s'%msg)
        return logon_service.logon(logon_dict, self.inst_path, self.nom_scribe)
        
    def remote_regedit(self, reg_liste):
        """Fonction appliquant les cls de registres dfinies dans regliste
           regliste = [{'type_variable' : CHAINE/DWORD, 'chemin' : y, 'variable' : z, 'valeur' : }, ...]
           EX. : reg_liste = { hkey : 'HKEY_...', <=  valer
                              keypath : 'Software\\Eole', 
                              keyname : 'test_key', 
                              keytype : 'REG_DWORD, <=  valer
                              value : 'SUPPR'} <= supprime la cl
        """
        self.log.info('Appel de la fonction remote_regedit "%s"'%reg_liste)
        try: return reg.put_regs(reg_liste)
        except Exception, e:
            self.log.error('%s'%e)
            self.log.debug('Erreur %s'%traceback.format_exc())

    def remote_execute(self, command, hide=False, nowait=True, waitinput=False, user=False):
        """excution  distance d'une ou d'une liste de commande(s)
        hide = cache, nowait = n'attend pas la sortie,
        waitinput = attend que le programme attende une entre (notepad par ex.)
        user = lancer la commande dans l'environnement utilisateur
        """
        # replacer WINDIR par self.windir
        self.log.debug('Appel')
        try:
            if type(command) == str or type(command) == unicode: command = [command]
            if user:
                pbgest = PBgest(ip='127.0.0.1', port=option.get_port_utilisateur())
                pbgest.d.addCallback(lambda object: object.callRemote('execute', cmd, hide, nowait, waitinput))
                return pbgest.d
            command = convert(command)
            for cmd in command:
                ret = lancecmd(cmd, hide, nowait, waitinput)
                self.log.info('Execution distante de : "%s", code retour : %s'%(cmd, ret))
            return True
        except Exception, e:
            self.log.error('%s'%e)
            self.log.debug('Erreur %s'%traceback.format_exc())

    def remote_killproc(self, progname):
        """Tue le programme progname
        """
        self.log.info('Appel de la fonction remote_killproc "%s"'%progname)
        try: return killProcName(progname)
        except Exception, e:
            self.log.error('%s'%e)
            self.log.debug('Erreur %s'%traceback.format_exc())


    def remote_winvnc(self, action, val=None, conf=None, restart=True):
        """gestion VNC, actions possibles:
            - start  - restart  - start_listen  - connect
            - stop              - stop_listen   - disconnect
            - setinputs (gestion clavier/souris)
        """
        self.log.info('Appel de la fonction remote_winvnc "%s"; val="%s" ; conf="%s"; restart="%s"'%(action, val, conf, restart))
        try:
            if action.lower() == 'start': self.vnc.start()
            if action.lower() == 'stop': self.vnc.stop()
            if action.lower() == 'restart': self.vnc.restart()
            if action.lower() == 'connect': self.vnc.connect(val)
            if action.lower() == 'disconnect': self.vnc.disconnect()
            if action.lower() == 'start_listen': self.vnc.start_user_listen(val, conf)
            if action.lower() == 'stop_listen': self.vnc.stop_listen()
            if action.lower() == 'setinputs': self.vnc.setinputs(val, restart)
            if action.lower() == 'set_type': self.vnc.set_type(val)
        except Exception, e:
            self.log.error('%s'%e)
            self.log.debug('Erreur %s'%traceback.format_exc())
        self.log.debug('Action %s terminee'%action)
        return True
    

    def remote_shutdown(self, reboot=0, force=False):
        """Extinction distante 0=teindre, 1=reboot, 2=fermeture de session
        """
        self.log.info('Appel de la fonction remote_shutdown "(reboot=%s, force=%s)"'%(reboot, force))
        #modification des privilges
        try:
            if int(reboot) == 0:
                opts = wc.EWX_POWEROFF
            elif int(reboot) == 1:
                opts = wc.EWX_REBOOT
            elif int(reboot) == 2:
                opts = wc.EWX_LOGOFF
            if force:
                opts = opts | wc.EWX_FORCE
            wa.ExitWindowsEx(opts, 0)
        except Exception, e:
            self.log.error('%s'%e)
            self.log.debug('Erreur %s'%traceback.format_exc())
        return True
##        try:
##            if int(reboot) in [0, 1]:
##                wa.InitiateSystemShutdown(None, '', 0, force, int(reboot))
##            elif int(reboot) == 2:
##                opts = wc.EWX_LOGOFF
##                if force: opts = opts | wc.EWX_FORCE
##                wa.ExitWindowsEx(opts,0)
##        except Exception, e:
##            self.log.error('%s'%e)
##            self.log.debug('Erreur %s'%traceback.format_exc())

    def remote_bloc(self, partmod, sid, logon=False):
        """Blocage de poste, mode partage (blocage internet fait par remote_execute(ipfw ...)
        connexion  utilisateur.exe qui va monter \\scribe\perso\devoirs et redmarrer explorer.exe
        pour prendre en compte les modif regedit.
        """
        self.log.info(u'Appel de la fonction remote_bloc : partmod=%s, sid=%s, logon=%s'%(partmod, sid, logon))
        try:
            factory = pb.PBClientFactory()
            reactor.connectTCP('127.0.0.1', self.user_port, factory)
            if int(partmod) == 2: #masque lecteurs et monte "devoirs"
                self.mask_part(sid, logon=logon)
                d = factory.getRootObject()
                d.addCallback(lambda object: object.callRemote("mount_devoir"))
            elif int(partmod) == 1: #masque lecteurs et NE monte PAS "devoirs" (le dmonte si besoin)
                self.mask_part(sid, logon=logon)
                d = factory.getRootObject()
                d.addCallback(lambda object: object.callRemote("umount_devoir"))
            elif int(partmod) == 0: #mode normal
                self.mask_part(sid, unmask=True, logon=logon)
            f = factory.getRootObject()
            if not logon: f.addCallback(lambda object: object.callRemote("restart_explorer"))
            return f
        except Exception, e:
            self.log.error('%s'%e)
            self.log.debug('Erreur %s'%traceback.format_exc())

        
    ######################
    # Fonctions internes #
    ######################
    def mask_part(self, sid, unmask=False, logon=False):
        """Masque les lecteurs S:, T:, U:
        """
        kpath = r'%s\Software\Microsoft\Windows\CurrentVersion\Policies\Explorer'%sid
        try: nodrives = int(reg.get_option(kpath, 'NoDrives', 'HKEY_USERS'))
        except: nodrives = 0
        try: noview = int(reg.get_option(kpath, 'NoViewOnDrive', 'HKEY_USERS'))
        except: noview = 0
        self.log.debug('logon=%s, nodrives=%s, noview=%s'%(logon, nodrives, noview))
        # sauvegarde le configuration d'origine des lecteurs
        if logon: self.nodrives, self.noview = nodrives, noview
        if unmask: # dsactivation
            nodrives, noview = self.nodrives, self.noview
        else: # activation mode devoir
            nodrives = lecteur.add_eole_drives(nodrives)
            noview = lecteur.add_eole_drives(noview)
        reg.addreg('HKEY_USERS', kpath, reg._winreg.REG_DWORD, 'NoDrives', nodrives)
        reg.addreg('HKEY_USERS', kpath, reg._winreg.REG_DWORD, 'NoViewOnDrive', noview)
        return True

class gestServ:
    def __init__(self, installSignalHandlers=0):
        self.installSignalHandlers = installSignalHandlers
        # l'adresse du Scribe
        self.serveur = option.get_ip_scribe()
        self.serveur_port = int(option.get_port_scribe())
        self.port = int(option.get_port_client())
        # Variables d'environnement
        self.tmpdir = os.path.join(os.environ['TEMP'], 'Eole')
        init_tmpdir(self.tmpdir)
        self.windir = os.environ['WINDIR'] 
        # on log sur 2 fichiers de 100ko
        handler = RotatingFileHandler(SERVICE_LOG,'a',100000,2)
        handler.setFormatter(FORMATTER)
        self.log = logging.getLogger()
        self.log.addHandler(handler)
        self.log.setLevel(option.get_log_level())

    #Fonctions de dmarrage et d'arrt du service
    def servstart(self):
        # affichage du dmarrage
        self.log.info("Le service Scribe demarre : ip Scribe=%s ; port Scribe=%s ; port local=%s ; windir=%s ; tmpdir=%s"
                                                %(self.serveur, self.serveur_port, self.port, self.windir, self.tmpdir))
        # connection au serveur
        factory = pb.PBClientFactory()
        reactor.connectTCP(self.serveur, self.serveur_port, factory)
        d = factory.getRootObject()
        computer_name = wa.GetComputerName()
        #appel  la mthode "remote_service_start"
        # (enregistrement de la station comme allume)
        d.addCallback(lambda object: object.callRemote("service_start", computer_name))
        #dmarrage du serveur local quelque soit la rponse du callRemote
        d.addErrback(self.servlisten)
        d.addCallback(self.servlisten)
        reactor.run(installSignalHandlers=self.installSignalHandlers)
    
    def servlisten(self, retval):
        self.log.info('retval : %s'%retval)
        # chargement des menus par dfaut
        reg.load_default_menus(self.tmpdir)
        fac = servFactory(Echoer(self), unsafeTracebacks=True)
        reactor.listenTCP(self.port, fac)

    def servstop(self):
        """Fonction d'arrt du service Scribe
        """
        # remise  l'origine des menu dmarrer et bureau communs
        # pour l'ouverture de sessions locales
        reg.load_default_menus(self.tmpdir)
        # connection au serveur
        factory = pb.PBClientFactory()
        reactor.connectTCP(self.serveur, 8789, factory)
        d = factory.getRootObject()
        #appel  la mthode "remote_service_stop"
        d.addCallback(lambda object: object.callRemote("service_stop"))
        #arrt du racteur  la rponse
        d.addErrback(lambda _: reactor.stop())
        d.addCallback(lambda _: reactor.stop())
    
def init_tmpdir(tmpdir):
    """supprime self.tmpdir s'il existe
    le re-cre
    """
    if os.path.exists(tmpdir):
        shutil.rmtree(tmpdir)
    os.makedirs(tmpdir)

